
Imports System.Collections.Generic
Imports System.IO

' ------------------------------------------------------------------ Writes a Track to a Stream.
Friend Class TrackWriter

    Private Shared ReadOnly TrackHeader As Byte() = {Asc("M"c), Asc("T"c), Asc("r"c), Asc("k"c)}

    ' -------------------------------------------------------------- the Track to write to the Stream
    Private m_track As New Track()

    ' -------------------------------------------------------------- the Stream to write to
    Private stream As Stream

    Private runningStatus As Integer = 0

    ' -------------------------------------------------------------- Track data in raw bytes
    Private trackData As New List(Of Byte)()

    Public Sub Write(ByVal strm As Stream)
        Me.stream = strm

        trackData.Clear()

        stream.Write(TrackHeader, 0, TrackHeader.Length)

        For Each e As MidiEvent In m_track.MidiEventIterator()
            WriteVariableLengthValue(e.DeltaTicks)

            Select Case e.MidiMessage.MessageType
                Case MessageType.Channel : Write(DirectCast(e.MidiMessage, ChannelMessage))

                Case MessageType.SystemExclusive : Write(DirectCast(e.MidiMessage, SysExMessage))

                Case MessageType.Meta : Write(DirectCast(e.MidiMessage, MetaMessage))

                Case MessageType.SystemCommon : Write(DirectCast(e.MidiMessage, SysCommonMessage))

                Case MessageType.SystemRealtime : Write(DirectCast(e.MidiMessage, SysRealtimeMessage))
            End Select
        Next

        Dim trackLength As Byte() = BitConverter.GetBytes(trackData.Count)

        If BitConverter.IsLittleEndian Then
            Array.Reverse(trackLength)
        End If

        stream.Write(trackLength, 0, trackLength.Length)

        For Each b As Byte In trackData
            stream.WriteByte(b)
        Next
    End Sub

    Private Sub WriteVariableLengthValue(ByVal value As Integer)
        Dim v As Integer = value
        Dim array As Byte() = New Byte(3) {}
        Dim count As Integer = 0

        array(0) = CByte((v And &H7F))

        v >>= 7

        While v > 0
            count += 1
            array(count) = CByte(((v And &H7F) Or &H80))
            v >>= 7
        End While

        While count >= 0
            trackData.Add(array(count))
            count -= 1
        End While
    End Sub

    Private Sub Write(ByVal message As ChannelMessage)
        If runningStatus <> message.Status Then
            trackData.Add(CByte(message.Status))
            runningStatus = message.Status
        End If

        trackData.Add(CByte(message.Data1))

        If ChannelMessage.DataBytesPerType(message.Command) = 2 Then
            trackData.Add(CByte(message.Data2))
        End If
    End Sub

    Private Sub Write(ByVal message As SysExMessage)
        ' --------------------------------------------------------- system exclusive message cancel running status
        runningStatus = 0

        trackData.Add(CByte(message.Status))

        WriteVariableLengthValue(message.Length - 1)

        For i As Integer = 1 To message.Length - 1
            trackData.Add(message(i))
        Next
    End Sub

    Private Sub Write(ByVal message As MetaMessage)
        trackData.Add(CByte(message.Status))
        trackData.Add(CByte(message.MetaType))

        WriteVariableLengthValue(message.Length)

        trackData.AddRange(message.GetBytes())
    End Sub

    Private Sub Write(ByVal message As SysCommonMessage)
        ' ------------------------------------------------------------- escaped message cancel running status
        runningStatus = 0
        ' ------------------------------------------------------------- Escaped message
        trackData.Add(CByte(&HF7))
        trackData.Add(CByte(message.Status))

        Select Case message.SysCommonType
            Case SysCommonType.MidiTimeCode
                trackData.Add(CByte(message.Data1))

            Case SysCommonType.SongPositionPointer
                trackData.Add(CByte(message.Data1))
                trackData.Add(CByte(message.Data2))

            Case SysCommonType.SongSelect
                trackData.Add(CByte(message.Data1))

        End Select
    End Sub

    Private Sub Write(ByVal message As SysRealtimeMessage)
        ' --------------------------------------------------------------- escaped messages cancel running status
        runningStatus = 0
        ' --------------------------------------------------------------- Escaped message
        trackData.Add(CByte(&HF7))
        trackData.Add(CByte(message.Status))
    End Sub

    ' -------------------------------------------------------------------- Get or set the Track to write to the Stream
    Public Property Track() As Track
        Get
            Return m_track
        End Get
        Set(ByVal value As Track)
            If value Is Nothing Then
                Throw New ArgumentNullException("Track")
            End If
            runningStatus = 0
            trackData.Clear()
            m_track = value
        End Set
    End Property

End Class
